@@ -47,19 +47,20 @@ nsUnknownContentTypeDialog.prototype = {
       this.mLauncher = aLauncher;
       this.mContext  = aContext;
       // Display the dialog using the Window Watcher interface.
+      
+      var ir = aContext.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
+      var dwi = ir.getInterface(Components.interfaces.nsIDOMWindowInternal);
       var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"]
                 .getService(Components.interfaces.nsIWindowWatcher);
-      this.mDialog = ww.openWindow(null, // no parent
+      this.mDialog = ww.openWindow(dwi,
                                    "chrome://mozapps/content/downloads/unknownContentType.xul",
                                    null,
-                                   "chrome,centerscreen,titlebar,dialog=yes",
+                                   "chrome,centerscreen,titlebar,dialog=yes,dependent",
                                    null);
       // Hook this object to the dialog.
       this.mDialog.dialog = this;
       
       // Hook up utility functions. 
-      // XXXben these lines can disappear if we can get XULPP to run on stuff not 
-      //        thus referenced in jar.mns. 
       this.getSpecialFolderKey = this.mDialog.getSpecialFolderKey;
       
       // Watch for error notifications.
@@ -90,7 +91,7 @@ nsUnknownContentTypeDialog.prototype = {
       var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
       try {
         result = prefs.getComplexValue(kDownloadFolderPref, Components.interfaces.nsILocalFile);
-        result = this.validateLeafName(result, aDefaultFile);
+        result = this.validateLeafName(result, aDefaultFile, aSuggestedFileExtension);
       }
       catch (e) { 
         // If we get here, it's because we have a new profile and the user has never configured download
@@ -98,9 +99,6 @@ nsUnknownContentTypeDialog.prototype = {
         // we need to discover the default save location. 
         var autodownload = prefs.getBoolPref("browser.download.useDownloadDir");
         if (autodownload) {
-          // XXXben we should really have a way of using XULPP on JS component files. 
-          // This code is very similar to that in pref-downloads.js, except the platform
-          // checks are done at runtime. ugh. 
           function getSpecialFolderKey(aFolderType) 
           {
             return "Home";
@@ -139,7 +137,7 @@ nsUnknownContentTypeDialog.prototype = {
           // code block. 
           prefs.setComplexValue("browser.download.defaultFolder", Components.interfaces.nsILocalFile, defaultFolder);
           
-          result = this.validateLeafName(defaultFolder, aDefaultFile);
+          result = this.validateLeafName(defaultFolder, aDefaultFile, aSuggestedFileExtension);
         }
       }
       
@@ -205,31 +203,56 @@ nsUnknownContentTypeDialog.prototype = {
       return result;
     },
     
-    validateLeafName: function (aLocalFile, aLeafName)
+    validateLeafName: function (aLocalFile, aLeafName, aFileExt)
     {
+      if (aLeafName == "")
+        aLeafName = "unnamed" + (aFileExt ? "." + aFileExt : "");
       aLocalFile.append(aLeafName);
-      
-      // Since we're automatically downloading, we don't get the file picker's 
-      // logic to check for existing files, so we need to do that here.
-      //
-      // Note - this code is identical to that in 
-      //   browser/base/content/contentAreaUtils.js. 
-      // If you are updating this code, update that code too! We can't share code
-      // here since this is called in a js component. 
-      while (aLocalFile.exists()) {
-        var parts = /.+-(\d+)(\..*)?$/.exec(aLocalFile.leafName);
-        if (parts) {
-          aLocalFile.leafName = aLocalFile.leafName.replace(/((\d+)\.)/, 
-                                                            function (str, p1, part, s) { 
-                                                              return (parseInt(part) + 1) + "."; 
-                                                            });
+
+      this.makeFileUnique(aLocalFile);
+
+      if (aLocalFile.isExecutable() && !this.mLauncher.targetFile.isExecutable()) {
+        var f = aLocalFile.clone();
+        aLocalFile.leafName = aLocalFile.leafName + "." + this.mLauncher.MIMEInfo.primaryExtension; 
+
+        f.remove(false);
+        this.makeFileUnique(aLocalFile);
+      }
+      return aLocalFile;
+    },
+
+    makeFileUnique: function (aLocalFile)
+    {
+      try {
+        // Since we're automatically downloading, we don't get the file picker's 
+        // logic to check for existing files, so we need to do that here.
+        //
+        // Note - this code is identical to that in 
+        //   browser/base/content/contentAreaUtils.js. 
+        // If you are updating this code, update that code too! We can't share code
+        // here since this is called in a js component. 
+        while (aLocalFile.exists()) {
+          var parts = /.+-(\d+)(\..*)?$/.exec(aLocalFile.leafName);
+          if (parts) {
+            aLocalFile.leafName = aLocalFile.leafName.replace(/((\d+)\.)|((\d+)$)/,
+                                                              function (str, dot, dotNum, noDot, noDotNum, pos, s) {
+                                                                return (parseInt(str) + 1) + (dot ? "." : "");
+                                                              });
+          }
+          else {
+            aLocalFile.leafName = aLocalFile.leafName.replace(/\.|$/, "-1$&");
+          }
         }
-        else {
-          aLocalFile.leafName = aLocalFile.leafName.replace(/\./, "-1$&");
+        aLocalFile.create(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);
+      }
+      catch (e) {
+        dump("*** exception in validateLeafName: " + e + "\n");
+        if (aLocalFile.leafName == "" || aLocalFile.isDirectory()) {
+          aLocalFile.append("unnamed");
+          if (aLocalFile.exists())
+            aLocalFile.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0600);
         }
       }
-      
-      return aLocalFile;
     },
     
     // ---------- implementation methods ----------
@@ -300,13 +323,14 @@ nsUnknownContentTypeDialog.prototype = {
 
       if (suggestedFileName)
         fname = suggestedFileName;
-        
+      
+      var displayName = fname.replace(/ +/g, " ");
 
-      this.mTitle = this.dialogElement("strings").getFormattedString("title", [fname]);
+      this.mTitle = this.dialogElement("strings").getFormattedString("title", [displayName]);
       win.setAttribute( "title", this.mTitle );
 
       // Put content type, filename and location into intro.
-      this.initIntro(url, fname);
+      this.initIntro(url, fname, displayName);
 
       var iconString = "moz-icon://" + fname + "?size=16&contentType=" + this.mLauncher.MIMEInfo.MIMEType;
       this.dialogElement("contentTypeImage").setAttribute("src", iconString);
@@ -319,7 +343,10 @@ nsUnknownContentTypeDialog.prototype = {
       // want users to be able to autodownload .exe files. 
       var rememberChoice = this.dialogElement("rememberChoice");
 
-      if (this.mLauncher.MIMEInfo.MIMEType == "application/octet-stream") {
+      var mimeType = this.mLauncher.MIMEInfo.MIMEType;
+      if (mimeType == "application/octet-stream" || 
+          mimeType == "application/x-msdownload" ||
+          this.mLauncher.targetFile.isExecutable()) {
         rememberChoice.checked = false;
         rememberChoice.disabled = true;
       }
@@ -328,12 +355,11 @@ nsUnknownContentTypeDialog.prototype = {
       }
       this.toggleRememberChoice(rememberChoice);
 
-
       // XXXben - menulist won't init properly, hack. 
       var openHandler = this.dialogElement("openHandler");
       openHandler.parentNode.removeChild(openHandler);
-      var openParent = this.dialogElement("open").parentNode;
-      openParent.appendChild(openHandler);
+      var openHandlerBox = this.dialogElement("openHandlerBox");
+      openHandlerBox.appendChild(openHandler);
 
       this.mDialog.setTimeout("dialog.postShowCallback()", 0);
     },
@@ -346,8 +372,9 @@ nsUnknownContentTypeDialog.prototype = {
     },
 
     // initIntro:
-    initIntro: function(url, filename) {
-        this.dialogElement( "location" ).value = filename;
+    initIntro: function(url, filename, displayname) {
+        this.dialogElement( "location" ).value = displayname;
+        this.dialogElement( "location" ).setAttribute("realname", filename);
 
         // if mSourcePath is a local file, then let's use the pretty path name instead of an ugly
         // url...
@@ -406,7 +433,7 @@ nsUnknownContentTypeDialog.prototype = {
             // On other platforms, default is Ok if there is a default app.
             // Note that nsIMIMEInfo providers need to ensure that this holds true
             // on each platform.
-        return this.mLauncher.MIMEInfo.defaultApplicationHandler;
+        return this.mLauncher.MIMEInfo.hasDefaultHandler;
     },
     
     // Set "default" application description field.
@@ -417,6 +444,12 @@ nsUnknownContentTypeDialog.prototype = {
         var defaultApp = this.dialogElement("strings").getFormattedString("defaultApp", [desc]);
         this.dialogElement("defaultHandler").label = defaultApp;
       }
+      else {
+        this.dialogElement("modeDeck").setAttribute("selectedIndex", "1");
+        // Hide the default handler item too, in case the user picks a 
+        // custom handler at a later date which triggers the menulist to show.
+        this.dialogElement("defaultHandler").hidden = true;
+      }
     },
 
     // getPath:
@@ -432,8 +465,10 @@ nsUnknownContentTypeDialog.prototype = {
       // from the browser at the moment because of security concerns. 
       var openWithDefaultOK = this.openWithDefaultOK();
       var mimeType = this.mLauncher.MIMEInfo.MIMEType;
-      if ((mimeType == "application/octet-stream" ||
-           mimeType == "application/x-msdownload") && !openWithDefaultOK) {
+      if (this.mLauncher.targetFile.isExecutable() || (
+          (mimeType == "application/octet-stream" ||
+           mimeType == "application/x-msdownload") && 
+           !openWithDefaultOK)) {
         this.dialogElement("open").disabled = true;
         var openHandler = this.dialogElement("openHandler");
         openHandler.disabled = true;
@@ -514,8 +549,11 @@ nsUnknownContentTypeDialog.prototype = {
     },
     
     openHandlerCommand: function () {
-      if (this.dialogElement("openHandler").selectedItem.id == "choose")
+      var openHandler = this.dialogElement("openHandler");
+      if (openHandler.selectedItem.id == "choose")
         this.chooseApp();
+      else
+        openHandler.setAttribute("lastSelectedItemID", openHandler.selectedItem.id);
     },
 
     updateOKButton: function() {
@@ -637,18 +675,7 @@ nsUnknownContentTypeDialog.prototype = {
           // for the file to be saved to to pass to |saveToDisk| - otherwise
           // we must ask the user to pick a save name.
 
-          var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
-          var targetFile = null;
-          try {
-            targetFile = prefs.getComplexValue("browser.download.defaultFolder", 
-                                               Components.interfaces.nsILocalFile);
-            var leafName = this.dialogElement("location").value;                                               
-            // Ensure that we don't overwrite any existing files here. 
-            targetFile = this.validateLeafName(targetFile, leafName);
-          }
-          catch(e) {}
-          
-          this.mLauncher.saveToDisk(targetFile, false);
+          this.mLauncher.saveToDisk(null, false);
         }
         else
           this.mLauncher.launchWithApplication(null, false);
@@ -703,6 +730,10 @@ nsUnknownContentTypeDialog.prototype = {
       fp.appendFilters(nsIFilePicker.filterApps);
 
       if (fp.show() == nsIFilePicker.returnOK && fp.file) {
+        // Show the "handler" menulist since we have a (user-specified) 
+        // application now.
+        this.dialogElement("modeDeck").setAttribute("selectedIndex", "0");
+        
         // Remember the file they chose to run.
         this.chosenApp = fp.file;
         // Update dialog.
@@ -711,9 +742,17 @@ nsUnknownContentTypeDialog.prototype = {
         otherHandler.setAttribute("path", this.getPath(this.chosenApp));
         otherHandler.label = this.chosenApp.leafName;
         this.dialogElement("openHandler").selectedIndex = 1;
+        this.dialogElement("openHandler").setAttribute("lastSelectedItemID", "otherHandler");
         
         this.dialogElement("mode").selectedItem = this.dialogElement("open");
       }
+      else {
+        var openHandler = this.dialogElement("openHandler");
+        var lastSelectedID = openHandler.getAttribute("lastSelectedItemID");
+        if (!lastSelectedID)
+          lastSelectedID = "defaultHandler";
+        openHandler.selectedItem = this.dialogElement(lastSelectedID);
+      }
     },
 
     // Turn this on to get debugging messages.
